/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.im.out.bpel.util;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.xml.namespace.QName;

import org.eclipse.bpel.model.Activity;
import org.eclipse.bpel.model.Assign;
import org.eclipse.bpel.model.BPELFactory;
import org.eclipse.bpel.model.BPELPackage;
import org.eclipse.bpel.model.Condition;
import org.eclipse.bpel.model.Copy;
import org.eclipse.bpel.model.Else;
import org.eclipse.bpel.model.ElseIf;
import org.eclipse.bpel.model.Exit;
import org.eclipse.bpel.model.Expression;
import org.eclipse.bpel.model.Flow;
import org.eclipse.bpel.model.From;
import org.eclipse.bpel.model.If;
import org.eclipse.bpel.model.Invoke;
import org.eclipse.bpel.model.PartnerActivity;
import org.eclipse.bpel.model.PartnerLink;
import org.eclipse.bpel.model.Process;
import org.eclipse.bpel.model.RepeatUntil;
import org.eclipse.bpel.model.Sequence;
import org.eclipse.bpel.model.To;
import org.eclipse.bpel.model.Variable;
import org.eclipse.bpel.model.Variables;
import org.eclipse.bpel.model.Wait;
import org.eclipse.bpel.model.While;
import org.eclipse.bpel.model.resource.BPELVariableResolver;
import org.eclipse.bpel.model.proxy.MessageProxy;
import org.eclipse.bpel.model.proxy.PartnerLinkProxy;
import org.eclipse.bpel.model.proxy.VariableProxy;
import org.eclipse.bpel.model.util.BPELUtils;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.wst.wsdl.Message;
import org.eclipse.bpel.model.resource.VariableResolver;
import org.eclipse.emf.common.util.EList;
import org.eclipse.stp.im.ServiceBinding;
import org.eclipse.stp.im.Step;

public class BPELGenerator {

	private Process process = null;
	private Resource bpelResource = null;
	private org.eclipse.stp.im.Process intermediateModel = null;
	private Step currentStep = null;
	
	public static VariableResolver VARIABLE_RESOLVER = new BPELVariableResolver();
	List<Runnable> fPass2Runnables = new ArrayList<Runnable>();

	public BPELGenerator (org.eclipse.stp.im.Process intermediateModel, Resource bpelResource) throws IOException {
		this.intermediateModel = intermediateModel;
		this.bpelResource = bpelResource;
		process = (Process)bpelResource.getContents().get(0);
	}

	public void generate() {
		// In pass 1 we parse and create the structural elements and attributes. 
		Process process = im2Process();
		if (process != null)
			bpelResource.getContents().set(0, process);
		// In pass 2, we run any postLoadRunnables which need to happen after
		pass2();
	}
	
	public Resource getResource() {
		return bpelResource;
	}
	
	/**
	 * In pass 2, we run any post load runnables which were queued during pass 1.
	 */
	protected void pass2() {	
		try {
			for(Runnable r : fPass2Runnables) {
				r.run();
			}
		} finally {			
			fPass2Runnables.clear();
		}
	}

	/**
	 * Sets a PartnerLink element for a given EObject. The given activity element
	 * must contain an attribute named "partnerLink".
	 * 
	 * @param activityStep  the imStep of the activity
	 * @param eObject  the EObject in which to set the partner link
	 */
	protected void setPartnerLink(Step activityStep, final EObject eObject, final EReference reference) {
		final String partnerLinkName = (String)activityStep.getProperties().get("partnerLink").getValue();
		if(partnerLinkName.equals(""))
			return;
		// We must do this as a post load runnable because the partner link might not
		// exist yet.
		fPass2Runnables.add(new Runnable() {
			public void run() {	
				PartnerLink targetPartnerLink = BPELUtils.getPartnerLink(eObject, partnerLinkName);
				if (targetPartnerLink == null) {
					targetPartnerLink = new PartnerLinkProxy(getResource().getURI(), partnerLinkName);
				}
				eObject.eSet(reference, targetPartnerLink);				
			}
		});		
	}

	/**
	 * Sets a Variable element for a given EObject. The given activity element
	 * must contain an attribute with the given name
	 * 
	 * @param activityStep  the imStep of the activity
	 * @param eObject  the EObject in which to set the variable
	 * @param variableAttrName  the name of the attribute containing the variable name
	 * @param reference  the EReference which is the variable pointer in EObject 
	 */
	protected void setVariable(Step activityStep, final EObject eObject, String variableNameAttr, final EReference reference) {
		final String variableName = (String)activityStep.getProperties().get(variableNameAttr).getValue();
		// We must do this as a post load runnable because the variable might not
		// exist yet.
		fPass2Runnables.add(new Runnable() {
			public void run() {				
				Variable targetVariable = getVariable(eObject, variableName);
				if (targetVariable == null) {
					targetVariable = new VariableProxy(getResource().getURI(), variableName);
				}		
				eObject.eSet(reference, targetVariable);				
			}
		});
  	} 
	
	protected void setVariable(final String variableName, final EObject eObject, final EReference reference) {
		// We must do this as a post load runnable because the variable might not
		// exist yet.
		fPass2Runnables.add(new Runnable() {
			public void run() {				
				Variable targetVariable = getVariable(eObject, variableName);
				if (targetVariable == null) {
					targetVariable = new VariableProxy(getResource().getURI(), variableName);
				}		
				eObject.eSet(reference, targetVariable);				
			}
		});
  	} 	

	/**
	 * Sets the standard attributes (name, joinCondition, and suppressJoinFailure).
	 */
	protected void setStandardAttributes(Step activityStep, Activity activity) {
		// Set name
		String name = activityStep.getName();
		activity.setName(name);

		// Set suppress join failure
		//Attr suppressJoinFailure = activityElement.getAttributeNode("suppressJoinFailure");
		//activity.setSuppressJoinFailure(new Boolean(suppressJoinFailure.getValue().equals("yes")));
	}


	/**
	 * Sets name, portType, operation, partner, variable and correlation for a given PartnerActivity object.
	 */
	protected void setOperationParms(final Step activityStep,
									 final PartnerActivity activity,
									 EReference variableReference,
									 EReference inputVariableReference,
									 EReference outputVariableReference,
									 EReference partnerReference) {

		// Set variable
		if (variableReference != null) {
			setVariable(activityStep, activity, "variable", variableReference);
		}
		if (inputVariableReference != null) {
			setVariable(activityStep, activity, "inputVariable", inputVariableReference);
		}
		if (outputVariableReference != null) {
			setVariable(activityStep, activity, "outputVariable", outputVariableReference);
		}

	}

	/**
	 * Converts an IM process to a BPEL Process object.
	 */
	@SuppressWarnings("nls")
	protected Process im2Process() {
		
		// Handle Variables Element
		List<org.eclipse.stp.im.Variable> variablesList = intermediateModel.getProcessVariables();
		if (variablesList != null) {
			process.setVariables(im2Variables(variablesList));
		}
		
		//the sequence activity was defined in the synch template
		Sequence sequenceActivity = (Sequence)process.getActivity();
		//activitiesList contains receive and reply defined in the template
		EList<Activity> activitiesList = sequenceActivity.getActivities();
		Step startStep = BpelUtil.getStartStep(intermediateModel.getSteps());
		//change first step name 
		String startStepName = startStep.getName();
		activitiesList.get(0).setName(startStepName);
		
		Activity replyActivity = activitiesList.remove(1);
		Activity initializeVariableAssign = generateInitAssign();
		activitiesList.add(initializeVariableAssign);
		currentStep = BpelUtil.getNextStep(startStep);
		while (!BpelUtil.isEndStep(currentStep)) {
			Activity activity = im2Activity(currentStep); 
			if (activity != null) //activity is null for flow and sequence 
				activitiesList.add(activity);
			currentStep = BpelUtil.getNextStep(currentStep);
		}
		replyActivity.setName(currentStep.getName());
		activitiesList.add(replyActivity);
		//xml2ExtensibleElement(process,processElement);
		
		return process;
	}
	
	protected Variables im2Variables(List<org.eclipse.stp.im.Variable> variablesList) {
		Variables variables = BPELFactory.eINSTANCE.createVariables();
		
		//generate the input and the output variables, actually non saved as im variables
		variables.getChildren().add(generateInputVariable());
		variables.getChildren().add(generateOutputVariable());
		for (org.eclipse.stp.im.Variable variable : variablesList) {
			variables.getChildren().add( im2Variable(variable) );
		}				
		
		return variables;
	}

  	protected Variable generateInputVariable() {
		Variable variable = BPELFactory.eINSTANCE.createVariable();
		variable.setName("input");
		QName qName = new QName(process.getTargetNamespace(), process.getName()+"RequestMessage");
		Message messageType = new MessageProxy(getResource().getURI(), qName);
		variable.setMessageType(messageType);
		return variable;
	}
	
	protected Variable generateOutputVariable() {
		Variable variable = BPELFactory.eINSTANCE.createVariable();
		variable.setName("output");
		QName qName = new QName(process.getTargetNamespace(), process.getName()+"ResponseMessage");
		Message messageType = new MessageProxy(getResource().getURI(), qName);
		variable.setMessageType(messageType);
		return variable;
	}
	
	/**
	 * Converts an IM variable Step to a BPEL Variable object.
	 */
	protected Variable im2Variable(org.eclipse.stp.im.Variable var) {
		Variable variable = BPELFactory.eINSTANCE.createVariable();
		
		// Set name
		if (var.getName() != null) 
			variable.setName(var.getName());
		
        return variable;
     }

     protected Activity im2Activity(Step activityStep) {
		Activity activity = null;
        
		String name = BpelUtil.getServiceBinding(activityStep).getName();        
        if (name.equals("BPEL-Receive")) {
       		//activity = im2Receive(activityServiceBinding);
     	} else if (name.equals("BPEL-Replay")) {
      		//activity = im2Reply(activityServiceBinding);
     	} else if (name.equals("BPEL-Invoke")) {
      		activity = im2Invoke(activityStep);
     	} else if (name.equals("BPEL-Assign")) {
      		activity = im2Assign(activityStep);
     	} else if (name.equals("BPEL-Wait")) {
      		activity = im2Wait(activityStep);
     	} else if (name.equals("BPEL-Sequence")) {
      		activity = im2Sequence(activityStep);
     	} else if (name.equals("BPEL-IfControl")) {
     		activity = im2If(activityStep);
     	} else if (name.equals("BPEL-Flow")) {
     		activity = im2Flow(activityStep);
     	} else if (name.equals("BPEL-Loop")) {
     		if(BpelUtil.executeAtLeastOne(activityStep))
     			activity = im2RepeatUntil(activityStep);
     		else
     			activity = im2While(activityStep);
     	} else {
     		return null;
     	}
     	  	
		return activity;
	}

	/**
	 * Converts an IM flow Step to a BPEL Flow object.
	 */
	protected Activity im2Flow(Step flowStep) {
		Flow flow = BPELFactory.eINSTANCE.createFlow();
		
        setStandardAttributes(flowStep, flow);
		ServiceBinding startFlowServiceBinding = null;
		Step step = null;
        for(Step startFlowStep : BpelUtil.getNextSteps(currentStep)) {
        	startFlowServiceBinding = BpelUtil.getServiceBinding(startFlowStep);
        	if(startFlowServiceBinding.getName().equals("BPEL-Sequence")) {
        		Activity activity = im2Activity(startFlowStep);
            	if (activity != null) {
            		flow.getActivities().add(activity);
            	}
        	} else {
        		Sequence sequence = BPELFactory.eINSTANCE.createSequence();
        		sequence.setName("supportSequence");
        		step = startFlowStep;
        		while (!BpelUtil.isEndFlowStep(step)) {
                       Activity activity = im2Activity(step);
                       if (activity != null) { 
                       		sequence.getActivities().add(activity);               	
                       }
                       step = BpelUtil.getNextStep(step);
                }
        		flow.getActivities().add(sequence);
        		
        	}
        }
        if (step != null)
        	this.currentStep = step; 
        return flow;
	}

	/**
	 * Converts an IM while Step to a BPEL While object.
	 */
	protected Activity im2While(Step whileStep) {
    	While _while = BPELFactory.eINSTANCE.createWhile();

    	org.eclipse.stp.im.Property xPathConditionProperty = whileStep.getProperties().get("XPathCondition");
		
		String condition = ( xPathConditionProperty != null)
							? (String)xPathConditionProperty.getValue()
							: "";
		// Handle condition
		Condition conditionBpel = im2Condition(condition);
		_while.setCondition(conditionBpel);
		
		Activity activity = null;
		Step firstStep = BpelUtil.getNextStep(whileStep);
		Step nextFirstStep = BpelUtil.getNextStep(firstStep);
		if (BpelUtil.isEndLoopStep(nextFirstStep, whileStep.getName())) {
			activity = im2Activity(firstStep);
			this.currentStep = nextFirstStep;
		}
		else
			activity = im2SupportSequence4Loop(whileStep);
		
		_while.setActivity(activity);		
		return _while;
	}

	/**
	 * Converts an IM if Step to a BPEL If object.
	 */
	protected Activity im2If(Step ifControlStep) {
		If _if = BPELFactory.eINSTANCE.createIf();		
		//_if.setElement(ifElement);

		List<Step> destinationsStepList = BpelUtil.getNextSteps(ifControlStep);
		Step ifStep = (Step)destinationsStepList.get(0);
		// Set activity
		Activity activity = null;
		Step nextIfStep = BpelUtil.getNextStep(ifStep);
		if (BpelUtil.isEndIfControlStep(nextIfStep)) {
			activity = im2Activity(ifStep);
			this.currentStep = nextIfStep;
		}
		else
			activity = im2SupportSequence4If(ifStep);
		if (activity != null) {
			_if.setActivity(activity);
		}
		
		String condition = BpelUtil.getCondition(ifControlStep, ifStep);
		// Handle condition element
		if (condition != null) {
			Condition conditionBpel = im2Condition(condition);
			_if.setCondition(conditionBpel);
		}
		
		Step elseStep = null;
		int indexDestinationsStepList;
		for (indexDestinationsStepList = 1; indexDestinationsStepList<destinationsStepList.size(); indexDestinationsStepList++) {
			elseStep = (Step)destinationsStepList.get(indexDestinationsStepList);
			condition = BpelUtil.getCondition(ifControlStep, elseStep);
			if (condition != null)
			//Handle elseif
				_if.getElseIf().add( im2ElseIf (elseStep, condition));
			else
			// Handle else
				_if.setElse(im2Else(elseStep));
		
		}
		
		setStandardAttributes(ifControlStep, _if);
		
		return _if;		
	}

	/**
	 * Converts an IM elseIf Step to a BPEL ElseIf object.
	 */
	protected ElseIf im2ElseIf(Step elseIfStep, String condition) {
		ElseIf elseIf = BPELFactory.eINSTANCE.createElseIf();
    	
		// Handle condition element
		if (condition != null) {
			Condition conditionBpel = im2Condition(condition);
			elseIf.setCondition(conditionBpel);
		}

		// Set activity
		Activity activity = null;
		Step nextElseIfStep = BpelUtil.getNextStep(elseIfStep);
		if (BpelUtil.isEndIfControlStep(nextElseIfStep)) {
			activity = im2Activity(elseIfStep);
			this.currentStep = nextElseIfStep;
		}
		else
			activity = im2SupportSequence4If(elseIfStep);
		
		elseIf.setActivity(activity);
		
		return elseIf;
	}

	protected Condition im2Condition(String imCondition) {
		Condition condition = BPELFactory.eINSTANCE.createCondition();
    	//im2Expression(imCondition, condition);
		condition.setBody(imCondition);
    	return condition;
	}
	
	protected Expression im2Expression(String imExpression) {
		Expression expression = BPELFactory.eINSTANCE.createExpression();
		expression.setBody(imExpression);
    	return expression;
	}

	protected Else im2Else(Step elseStep) {
		Else _else = BPELFactory.eINSTANCE.createElse();
		
		Activity activity = null;
		Step nextElseStep = BpelUtil.getNextStep(elseStep);
		if (BpelUtil.isEndIfControlStep(nextElseStep)) {
			activity = im2Activity(elseStep);
			this.currentStep = nextElseStep;
		}
		else
			activity = im2SupportSequence4If(elseStep);
		
		_else.setActivity(activity);
		
		return _else;
	}


	/**
	 * Converts an IM sequence Step to a BPEL Sequence object.
	 */
	protected Activity im2Sequence(Step sequenceStep) {
		Sequence sequence = BPELFactory.eINSTANCE.createSequence();
		Step step = BpelUtil.getNextStep(sequenceStep);
		while (!BpelUtil.isEndSequenceStep(step)) {
               Activity activity = im2Activity(step);
               if (activity != null) { 
               		sequence.getActivities().add(activity);               	
               }
               step = BpelUtil.getNextStep(step);
        }
		this.currentStep = step;
        setStandardAttributes(sequenceStep, sequence);
		return sequence;
	} 
	
	protected Activity im2SupportSequence4If(Step ifStep) {
		Sequence sequence = BPELFactory.eINSTANCE.createSequence();
		Step step = ifStep;
		while (!BpelUtil.isEndIfControlStep(step)) {
               Activity activity = im2Activity(step);
               if (activity != null) { 
               		sequence.getActivities().add(activity);               	
               }
               step = BpelUtil.getNextStep(step);
        }
		this.currentStep = step;
        setStandardAttributes(ifStep, sequence);
		return sequence;
	} 
	
	protected Activity im2SupportSequence4Loop(Step loopStep) {
		Sequence sequence = BPELFactory.eINSTANCE.createSequence();
		Step step = BpelUtil.getNextStep(loopStep);
		while (!BpelUtil.isEndLoopStep(step, loopStep.getName())) {
               Activity activity = im2Activity(step);
               if (activity != null) { 
               		sequence.getActivities().add(activity);               	
               }
               step = BpelUtil.getNextStep(step);
        }
		this.currentStep = step;
        setStandardAttributes(loopStep, sequence);
		return sequence;
	} 

	/**
	 * Converts an IM wait Step to a BPEL Wait object.
	 */
	protected Activity im2Wait(Step waitStep) {
    	Wait wait = BPELFactory.eINSTANCE.createWait();
		
		
    	org.eclipse.stp.im.Property modeProperty = waitStep.getProperties().get("mode");
    	if (  modeProperty != null) {
    		String valueWaiting = (waitStep.getProperties().get("value") != null) 
    								? (String)waitStep.getProperties().get("value").getValue()
    								: "";
    		// Set until 
    		if(((String)modeProperty.getValue()).equalsIgnoreCase("Date")) 
    			wait.setUntil(im2Expression(valueWaiting));
    		else //Set for
    			wait.setFor(im2Expression(valueWaiting));
    		
    	}
		
		setStandardAttributes(waitStep, wait);
			
		return wait;						
	}

	/**
	 * Converts a IM Step exit to a BPEL Exit object.
	 */
	protected Activity im2Exit(Step exitStep) {
    	Exit exit = BPELFactory.eINSTANCE.createExit();
		setStandardAttributes(exitStep, exit);
			
		return exit;
	}
	
	protected Activity generateInitAssign() {
		Assign assign = BPELFactory.eINSTANCE.createAssign();
		assign.setValidate(new Boolean(true));
		assign.setName("InitializeVariable");
		
		Copy copy = BPELFactory.eINSTANCE.createCopy();
		
		//generate literal from
        From from = BPELFactory.eINSTANCE.createFrom(); 
        StringBuilder elementData = new StringBuilder(256);
        elementData.append("<xsd:" + process.getName() + "Response xmlns:xsd=\"" + process.getTargetNamespace() + "\">");
        elementData.append("<xsd:result/>");
        elementData.append("</xsd:" + process.getName() + "Response>");
        from.setUnsafeLiteral(Boolean.FALSE);
		String elementDataFinal = elementData.toString();			
		if (isEmptyOrWhitespace(elementDataFinal) == false) {
			from.setUnsafeLiteral(Boolean.TRUE);
			from.setLiteral( elementDataFinal );
		}
        copy.setFrom(from);
        
        //generate to
        To to = BPELFactory.eINSTANCE.createTo();
        setVariable("output", to, BPELPackage.eINSTANCE.getTo_Variable());
//        ((ToImpl)to).setPartName("payload");
        Expression expressionObject = BPELFactory.eINSTANCE.createExpression();
		to.setExpression(expressionObject);
        copy.setTo(to);

        assign.getCopy().add(copy);
		return assign;
	}

	/**
	 * Converts an IM assign step to a BPEL Assign object.
	 */
	protected Activity im2Assign(Step assignStep) {
		Assign assign = BPELFactory.eINSTANCE.createAssign();		
       
        assign.getCopy().add( im2Copy ( assignStep ));
        
        setStandardAttributes(assignStep, assign);
		return assign;
	}

	/**
	 * Converts an IM copy step to a BPEL Copy object.
	 */
	protected Copy im2Copy(Step copyStep) {
		Copy copy = BPELFactory.eINSTANCE.createCopy();

		String fromVariable = (copyStep.getProperties().get("fromVariable") != null) 
								? ((String)copyStep.getProperties().get("fromVariable").getValue())
								: "";
		String fromConstant = (copyStep.getProperties().get("fromConstant") != null) 
								? ((String)copyStep.getProperties().get("fromConstant").getValue())
								: "";
        if (!fromVariable.equals("") || !fromConstant.equals("")) {
            From from = BPELFactory.eINSTANCE.createFrom(); 
            im2From(from, copyStep); 
            copy.setFrom(from);
        }
        
        String toVariable = (String)copyStep.getProperties().get("to").getValue();
        if (!toVariable.equals("")) {
            To to = BPELFactory.eINSTANCE.createTo();
            im2To(to, copyStep); 
            copy.setTo(to);
        }
        
		return copy;
	}

	/**
	 * Converts an IM "to" step to a BPEL To object.
	 */
	protected void im2To(To to, Step toStep) {
		// Set variable
		String variable = (String)toStep.getProperties().get("to").getValue(); 
    
		if (!variable.equals("")) 				
			setVariable(toStep, to, "to", BPELPackage.eINSTANCE.getTo_Variable());
			
		// must be expression
		Expression expressionObject = BPELFactory.eINSTANCE.createExpression();
		to.setExpression(expressionObject);
	}

	/**
	 * Converts an IM "from" step to a BPEL From object.
	 */
	protected void im2From(From from, Step fromStep) {
		// Set variable
		String assignmentType = (String)fromStep.getProperties().get("assignmentType").getValue();
		if (assignmentType.equals(IConstants.variableValueInAssign)) {
			String variable = (String)fromStep.getProperties().get("fromVariable").getValue();
			if (!variable.equals("")) 			
				setVariable(fromStep, from, "fromVariable", BPELPackage.eINSTANCE.getFrom_Variable() );
	
			// must be expression
			Expression expressionObject = BPELFactory.eINSTANCE.createExpression();
			from.setExpression(expressionObject);
		} else if (assignmentType.equals(IConstants.fixedValueInAssign)) {
			String constantValue = (String)fromStep.getProperties().get("fromConstant").getValue();
			from.setUnsafeLiteral(Boolean.FALSE);
			if (isEmptyOrWhitespace(constantValue) == false) {
				from.setUnsafeLiteral(Boolean.TRUE);
				from.setLiteral( constantValue );
			}
		} else if (assignmentType.equals(IConstants.expressionValueInAssign)) {
			String expressionValue = (String)fromStep.getProperties().get("fromExpression").getValue();
			Expression expression = im2Expression(expressionValue);
			from.setExpression(expression);
		}
	}

	/**
	 * Converts an IM invoke step to a BPEL Invoke object.
	 */
	protected Activity im2Invoke(Step invokeStep) {
		Invoke invoke = BPELFactory.eINSTANCE.createInvoke();
		
		// Set several parms
		setStandardAttributes(invokeStep, invoke);
		setOperationParms(invokeStep, invoke, null, BPELPackage.eINSTANCE.getInvoke_InputVariable(), BPELPackage.eINSTANCE.getInvoke_OutputVariable(), BPELPackage.eINSTANCE.getPartnerActivity_PartnerLink());

		return invoke;
	}

	/**
	 * Converts an repeatUntil imStep to a BPEL RepeatUntil object.
	 */
	protected Activity im2RepeatUntil(Step repeatUntilStep) {
		RepeatUntil repeatUntil = BPELFactory.eINSTANCE.createRepeatUntil();
		
		setStandardAttributes(repeatUntilStep, repeatUntil);

		org.eclipse.stp.im.Property xPathConditionProperty = repeatUntilStep.getProperties().get("XPathCondition");
		
		String condition = ( xPathConditionProperty != null)
							? (String)xPathConditionProperty.getValue()
							: "";
		// Handle condition
		Condition conditionBpel = im2Condition(condition);
		repeatUntil.setCondition(conditionBpel);
		
		Activity activity = null;
		Step firstStep = BpelUtil.getNextStep(repeatUntilStep);
		Step nextFirstStep = BpelUtil.getNextStep(firstStep);
		if (BpelUtil.isEndLoopStep(nextFirstStep, repeatUntilStep.getName())) {
			activity = im2Activity(firstStep);
			this.currentStep = nextFirstStep;
		}
		else
			activity = im2SupportSequence4Loop(repeatUntilStep);
		
		repeatUntil.setActivity(activity);
		
		return repeatUntil;
	}

	/**
     * Returns true if the string is either null or contains just whitespace.
	 * @param value 
	 * @return true if empty or whitespace, false otherwise.
     */
		
   static public boolean isEmptyOrWhitespace( String value )
   {
       if( value == null || value.length() == 0) {
           return true;
       }               
       for( int i = 0, j = value.length(); i < j; i++ )
       {
           if( ! Character.isWhitespace( value.charAt(i) ) ) {
               return false;
           }
       }
       return true;
   }
	
	/**
	 * @param eObject
	 * @param variableName
	 * @return the resolved variable
	 */
	public static Variable getVariable(EObject eObject, String variableName) {
		return VARIABLE_RESOLVER.getVariable(eObject, variableName);
	}	
}
