/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.im.out.metadb.processor;

import java.io.File;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.spagic.im.out.metadb.IM2MetaDBActivator;
import org.spagic.utils.SpagicLogger;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Manages the binding specific processors
 *
 */
public class RuntimeProcessor {
	
	private List<IProcessor> processors = null;
	private static RuntimeProcessor instance = null;
	
    public static RuntimeProcessor getInstance() {
        if (instance == null) {
            synchronized (RuntimeProcessor.class) {
                if (instance == null) {
                    try {
                        instance = new RuntimeProcessor();
                    } // try
                    catch (Exception ex) {
                        System.out
                                .println("RuntimeProcessor::getInstance: error creating instance");
                        ex.printStackTrace();
                    } // catch(Exception ex) try
                } // if (instance == null)
            } // synchronized(RuntimeProcessor.class)
        } // if (instance == null)
        return instance;
    } // public static RuntimeProcessor getInstance()

    private RuntimeProcessor() {
		super();
		init(IM2MetaDBActivator.getFileFromPlugin("conf/processors.xml"));
	}
	
	/**
	 * Read the configuration file and initializes the processors.
	 * @param f The configuration file
	 */
	private void init(File f) {
		try {
			// Parse the XML as a W3C document.
			DocumentBuilder builder = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder();
			Document document = builder.parse(f);
			initProcessors(document);
		} catch (Exception ex) {
			SpagicLogger.error(IM2MetaDBActivator.PLUGIN_ID,"Error reading configuration file", ex);
		}
	}
	
	/**
	 * Instantiates all binding specific processors
	 * @param document Processors configuration DOM
	 * @throws Exception
	 */
	private void initProcessors(Document document) throws Exception {
		String xPath ="/runtime/metadb-processors/metadb-processor";
		NodeList nodes = selectNodes(document, xPath); 
		
		this.processors = new ArrayList<IProcessor>();
		IProcessor aProcessor = null;
		String name = null;
		String binding = null;
		String classQName = null;
		String technology = null;
		Node n = null;
		for (int i=0; i<nodes.getLength(); i++) {
			n = nodes.item(i);
			name = getAttributeValue(n, "name");
			binding = getAttributeValue(n, "binding");
			classQName = getAttributeValue(n, "class");
			technology = getAttributeValue(n, "technology");
			
			Class<IProcessor> processorClass = (Class<IProcessor>)Class.forName(classQName);
			Constructor<IProcessor> constructor = processorClass.getDeclaredConstructor(String.class, String.class, String.class);
			aProcessor = constructor.newInstance(new Object[]{name, binding, technology});
			this.processors.add(aProcessor);
		} 
	
	}
	
	private NodeList selectNodes(Node document, String xPath) throws XPathExpressionException {
		XPath xpath = XPathFactory.newInstance().newXPath();
        return (NodeList)xpath.evaluate(xPath, document, XPathConstants.NODESET);
	}
	
	private String getAttributeValue(Node n, String attributeName){
		try{
			Node attributeNode = n.getAttributes().getNamedItem(attributeName); 
			return (attributeNode != null ? attributeNode.getNodeValue() : "");
		}catch (Exception e){
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Retrieves a processor for a specific binding.
	 * The processor should be thread safe because the same instance could be invoked
	 * in multithread.
	 * @param binding Binding identifier
	 * @return The processor specific for the specified binding.
	 */
	public IProcessor getProcessor(String runtime) {
		for (IProcessor processor : processors) {
			if (processor.getRuntime().equals(runtime)) {
				return processor;
			}
		}
		return null;
	}

}
