/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.im.out.metadb.processor.impl;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.stp.im.ImFactory;
import org.eclipse.stp.im.ObservableAttribute;
import org.eclipse.stp.im.ServiceBinding;
import org.eclipse.stp.im.Step;
import org.eclipse.stp.im.StpIntermediateModel;
import org.eclipse.stp.im.Transition;
import org.spagic.im.out.metadb.IM2MetaDBActivator;
import org.spagic.utils.SpagicLogger;
import org.spagic.utils.SpagicUtils;


/**
 * ODE BPEL preprocessor
 *
 */
public class BPELProcessor extends BaseProcessor {

	private static final String START_SERVICE = "StartService";
	private static final String RECEIVE_OPERATION = "BPEL-Receive";
	
	private static final String ASSIGN_NAME = "InitializeVariable";
	private static final String ASSIGN_DESCRIPTION = "Step for initialize all process variables";
	
	public BPELProcessor(String name, String runtime, String technology) {
		super(name, runtime, technology);
	}

	public org.eclipse.stp.im.Process preProcess(org.eclipse.stp.im.Process process, StpIntermediateModel stpIntermediateModel) {
		// List for storing new steps and transition to the end of
		// preprocessing to avoid ConcurrentModificationException
		List<Step> newStepsList = new ArrayList<Step>();

		ImFactory stpModelfactory = ImFactory.eINSTANCE;

		List<Step> stepList = process.getSteps();
		for (Step step : stepList) {
			// Retrieve the suitable binding for JBI
			ServiceBinding serviceBinding = SpagicUtils.getRightStepBinding(step, getRuntime());
			
			if (serviceBinding == null) {
				SpagicLogger.error(IM2MetaDBActivator.PLUGIN_ID,"Unable to retrieve service binding for runtime " + getRuntime(), null);
				return process;
			}
			
			
			// Check if the binding is related to a special case
			// Find the first process step (in the Intermediate Model the steps are not ordered)
			if (step.getServiceModel().getServiceName().equals(START_SERVICE) &&
				serviceBinding.getName().equals(RECEIVE_OPERATION)) {
				
				// Retrieve transition from first step to next step
				List<Transition> transitionList = step.getSourceTransitions();
				if (transitionList.size() > 1) {
					SpagicLogger.error(IM2MetaDBActivator.PLUGIN_ID, "Too much transitions starting from " + step.getName() + " step", null);
					return process;
				}
				Transition firstTransition = transitionList.get(0);
				Step secondStep = firstTransition.getTarget();
				
				// Remove the transition between first step and second step
				process.getTransitions().remove(firstTransition);
				
				
				
				// Create a new component for the variable initialization
				Step initializeVariablesAssignStep = stpModelfactory.createStep();
				initializeVariablesAssignStep.setName(ASSIGN_NAME);
				initializeVariablesAssignStep.setDescription(ASSIGN_DESCRIPTION);
				
				// Create two transition (in and out) from the Enricher step to the new component
				Transition inTransition = stpModelfactory.createTransition();
				Transition outTransition = stpModelfactory.createTransition();
				
				inTransition.setSource(step);
				inTransition.setTarget(initializeVariablesAssignStep);
				
				outTransition.setSource(initializeVariablesAssignStep);
				outTransition.setTarget(secondStep);
				
				// Add the new step and transitions to the process
				newStepsList.add(initializeVariablesAssignStep);
				process.getTransitions().add(inTransition);
				process.getTransitions().add(outTransition);	
				
				// Check if secondStep has some relevant data associated, because they must be moved
				// on the assign step to ensure proper work of iter management 
				List<ObservableAttribute> secondStepObservableAttributes = secondStep.getObservableAttributes();
				// Move all observable attributes to InitializeAttributes step
				initializeVariablesAssignStep.getObservableAttributes().addAll(secondStepObservableAttributes);
				secondStepObservableAttributes.clear();
				
				break;
			}
		}
		
		// Modify the IM model (cannot modify directly in the previous for because the
		// iteration fails)
		for (Step step : newStepsList) {
			process.getSteps().add(step);
		}
				
		// Add a fixed step for initializing all variables 
		return process;
	}

}
